from os import getenv
from contextlib import contextmanager
from sqlalchemy import create_engine
from sqlalchemy.orm import sessionmaker, declarative_base

from logdetective import logger


def get_pg_url() -> str:
    """create postgresql connection string"""
    return (
        f"postgresql+psycopg2://{getenv('POSTGRESQL_USER')}"
        f":{getenv('POSTGRESQL_PASSWORD')}@{getenv('POSTGRESQL_HOST', 'postgres')}"
        f":{getenv('POSTGRESQL_PORT', '5432')}/{getenv('POSTGRESQL_DATABASE')}"
    )


# To log SQL statements, set SQLALCHEMY_ECHO env. var. to True|T|Yes|Y|1
sqlalchemy_echo = getenv("SQLALCHEMY_ECHO", "False").lower() in (
    "true",
    "t",
    "yes",
    "y",
    "1",
)
engine = create_engine(get_pg_url(), echo=sqlalchemy_echo)
SessionFactory = sessionmaker(autoflush=True, bind=engine)
Base = declarative_base()


@contextmanager
def transaction(commit: bool = False):
    """
    Context manager for 'framing' a db transaction.

    Args:
        commit: Whether to call `Session.commit()` upon exiting the context. Should be set to True
            if any changes are made within the context. Defaults to False.
    """

    session = SessionFactory()
    try:
        yield session
        if commit:
            session.commit()
    except Exception as ex:
        logger.warning("Exception while working with database: %s", str(ex))
        session.rollback()
        raise
    finally:
        session.close()


def init():
    """Init db"""
    Base.metadata.create_all(engine)
    logger.debug("Database initialized")


def destroy():
    """Destroy db"""
    Base.metadata.drop_all(engine)
    logger.warning("Database cleaned")


DB_MAX_RETRIES = 3  # How many times retry a db operation



from logdetective.server.database.base import Base
from logdetective.server.database.models.merge_request_jobs import (
    Forge,
    GitlabMergeRequestJobs,
    Comments,
    Reactions,
)
from logdetective.server.database.models.koji import (
    KojiTaskAnalysis,
)
from logdetective.server.database.models.metrics import (
    AnalyzeRequestMetrics,
    EndpointType,
)
from logdetective.server.database.models.exceptions import (
    KojiTaskNotFoundError,
    KojiTaskNotAnalyzedError,
    KojiTaskAnalysisTimeoutError,
)

__all__ = [
    Base.__name__,
    GitlabMergeRequestJobs.__name__,
    Comments.__name__,
    Reactions.__name__,
    AnalyzeRequestMetrics.__name__,
    EndpointType.__name__,
    Forge.__name__,
    KojiTaskAnalysis.__name__,
    KojiTaskNotFoundError.__name__,
    KojiTaskNotAnalyzedError.__name__,
    KojiTaskAnalysisTimeoutError.__name__,
]



import logging
from urllib.parse import urlparse

import aiohttp
from aiohttp.web import HTTPBadRequest

LOG = logging.getLogger("logdetective")


class RemoteLog:
    """
    Handles retrieval of remote log files.
    """

    def __init__(self, url: str, http_session: aiohttp.ClientSession):
        """
        Initialize with a remote log URL and HTTP session.

        Args:
            url: A remote URL pointing to a log file
            http_session: The HTTP session used to retrieve the remote file
        """
        self._url = url
        self._http_session = http_session

    @property
    def url(self) -> str:
        """The remote log url."""
        return self._url

    @property
    async def content(self) -> str:
        """Content of the url."""
        return await self.get_url_content()

    def validate_url(self) -> bool:
        """Validate incoming URL to be at least somewhat sensible for log files
        Only http and https protocols permitted. No result, params or query fields allowed.
        Either netloc or path must have non-zero length.
        """
        result = urlparse(self.url)
        if result.scheme not in ["http", "https"]:
            return False
        if any([result.params, result.query, result.fragment]):
            return False
        if not (result.path or result.netloc):
            return False
        return True

    async def get_url_content(self) -> str:
        """validate log url and return log text."""
        if self.validate_url():
            LOG.debug("process url %s", self.url)
            try:
                response = await self._http_session.get(self.url, raise_for_status=True)
            except (aiohttp.ClientResponseError, aiohttp.ClientConnectorError) as ex:
                raise RuntimeError(f"We couldn't obtain the logs: {ex}") from ex
            return await response.text()
        LOG.error("Invalid URL received ")
        raise RuntimeError(f"Invalid log URL: {self.url}")

    async def process_url(self) -> str:
        """Validate log URL and return log text."""
        try:
            return await self.get_url_content()
        except RuntimeError as ex:
            raise HTTPBadRequest(reason=f"We couldn't obtain the logs: {ex}") from ex



import asyncio
import re
from typing import Any, Callable, Optional

import backoff
import koji
from logdetective.server.config import LOG
from logdetective.server.exceptions import (
    KojiInvalidTaskID,
    LogDetectiveConnectionError,
    LogsMissingError,
    LogsTooLargeError,
    UnknownTaskType,
)


FAILURE_LOG_REGEX = re.compile(r"(\w*\.log)")


def connection_error_giveup(details: backoff._typing.Details) -> None:
    """
    Too many connection errors, give up.
    """
    LOG.error("Too many connection errors, giving up. %s", details["exception"])
    raise LogDetectiveConnectionError() from details["exception"]


@backoff.on_exception(
    backoff.expo,
    koji.GenericError,
    max_time=60,
)
async def call_koji(func: Callable, *args, **kwargs) -> Any:
    """
    Call a Koji function asynchronously.

    Automatically retries on connection errors.
    """
    try:
        result = await asyncio.to_thread(func, *args, **kwargs)
    except koji.ActionNotAllowed as e:
        # User doesn't have permission to do this, don't retry.
        raise LogDetectiveConnectionError(e) from e
    return result


async def get_failed_subtask_info(
    koji_session: koji.ClientSession, task_id: int
) -> dict[str, Any]:
    """
    If the provided task ID represents a task of type "build", this function
    will return the buildArch or rebuildSRPM subtask that failed. If there is
    more than one, it will return the first one found from the following
    ordered list of processor architectures. If none is found among those
    architectures, it will return the first failed architecture after a
    standard sort.
    * x86_64
    * aarch64
    * riscv
    * ppc64le
    * s390x

    If the provided task ID represents a task of type "buildArch" or
    "buildSRPMFromSCM" and has a task state of "FAILED", it will be returned
    directly.

    Any other task type will rase the UnknownTaskType exception.

    If no task or subtask of the provided task is in the task state "FAILED",
    this function will raise a NoFailedSubtask exception.
    """

    # Look up the current task first and check its type.
    taskinfo = await call_koji(koji_session.getTaskInfo, task_id)
    if not taskinfo:
        raise KojiInvalidTaskID(f"Task {task_id} does not exist.")

    # If the parent isn't FAILED, the children probably aren't either.
    # There's one special case where the user may have canceled the
    # overall task when one arch failed, so we should check that situation
    # too.
    if (
        taskinfo["state"] != koji.TASK_STATES["FAILED"]
        and taskinfo["state"] != koji.TASK_STATES["CANCELED"]  # noqa: W503 flake vs lint
    ):
        raise UnknownTaskType(f"The primary task state was {taskinfo['state']}.")

    # If the task is buildArch or buildSRPMFromSCM, we can return it directly.
    if taskinfo["method"] in ["buildArch", "buildSRPMFromSCM"]:
        return taskinfo

    # Look up the subtasks for the task.
    response = await asyncio.to_thread(koji_session.getTaskDescendents, task_id)
    subtasks = response[f"{task_id}"]
    arch_tasks = {}
    for subtask in subtasks:
        if (
            subtask["method"] not in ["buildArch", "buildSRPMFromSCM"]
            or subtask["state"] != koji.TASK_STATES["FAILED"]  # noqa: W503 flake vs lint
        ):
            # Skip over any completed subtasks or non-build types
            continue

        arch_tasks[subtask["arch"]] = subtask

    # Return the first architecture in the order of preference.
    for arch in ["x86_64", "aarch64", "riscv", "ppc64le", "s390x"]:
        if arch in arch_tasks:
            return arch_tasks[arch]

    # If none of those architectures were found, return the first one
    # alphabetically
    return arch_tasks[sorted(arch_tasks.keys())[0]]


async def get_failed_log_from_task(
    koji_session: koji.ClientSession, task_id: int, max_size: int
) -> Optional[tuple[str, str]]:
    """
    Get the failed log from a task.

    If the log is too large, this function will raise a LogsTooLargeError.
    If the log is missing or garbage-collected, this function will raise a
    LogsMissingError.
    """
    taskinfo = await get_failed_subtask_info(koji_session, task_id)

    # Read the failure reason from the task. Note that the taskinfo returned
    # above may not be the same as passed in, so we need to use taskinfo["id"]
    # to look up the correct failure reason.
    result = await call_koji(
        koji_session.getTaskResult, taskinfo["id"], raise_fault=False
    )

    # Examine the result message for the appropriate log file.
    match = FAILURE_LOG_REGEX.search(result["faultString"])
    if match:
        failure_log_name = match.group(1)
    else:
        # The best thing we can do at this point is return the
        # task_failed.log, since it will probably contain the most
        # relevant information
        return result["faultString"]

    # Check that the size of the log file is not enormous
    task_output = await call_koji(
        koji_session.listTaskOutput, taskinfo["id"], stat=True
    )
    if not task_output:
        # If the task has been garbage-collected, the task output will be empty
        raise LogsMissingError(
            "No logs attached to this task. Possibly garbage-collected."
        )

    if failure_log_name not in task_output:
        # This shouldn't be possible, but we'll check anyway.
        raise LogsMissingError(f"{failure_log_name} could not be located")

    if int(task_output[failure_log_name]["st_size"]) > max_size:
        raise LogsTooLargeError(
            f"{task_output[failure_log_name]['st_size']} exceeds max size {max_size}"
        )

    log_contents = await call_koji(
        koji_session.downloadTaskOutput, taskinfo["id"], failure_log_name
    )
    return failure_log_name, log_contents.decode("utf-8")



from datetime import datetime, timedelta, timezone
from sqlalchemy import Column, BigInteger, DateTime, ForeignKey, Integer, String
from sqlalchemy.orm import relationship
from sqlalchemy.exc import OperationalError
import backoff

from logdetective.server.config import SERVER_CONFIG
from logdetective.server.compressors import LLMResponseCompressor
from logdetective.server.database.models.metrics import AnalyzeRequestMetrics
from logdetective.server.database.base import Base, transaction, DB_MAX_RETRIES
from logdetective.server.database.models.exceptions import (
    KojiTaskNotFoundError,
    KojiTaskNotAnalyzedError,
    KojiTaskAnalysisTimeoutError,
)
from logdetective.server.models import KojiStagedResponse


class KojiTaskAnalysis(Base):
    """Store details for the koji task analysis"""

    __tablename__ = "koji_task_analysis"

    id = Column(Integer, primary_key=True)
    koji_instance = Column(String(255), nullable=False, index=True)
    task_id = Column(BigInteger, nullable=False, index=True, unique=True)
    log_file_name = Column(String(255), nullable=False, index=True)
    request_received_at = Column(
        DateTime,
        nullable=False,
        index=True,
        default=datetime.now(timezone.utc),
        comment="Timestamp when the request was received",
    )
    response_id = Column(
        Integer,
        ForeignKey("analyze_request_metrics.id"),
        nullable=True,
        index=False,
        comment="The id of the analyze request metrics for this task",
    )
    response = relationship("AnalyzeRequestMetrics")

    @classmethod
    @backoff.on_exception(backoff.expo, OperationalError, max_tries=DB_MAX_RETRIES)
    def create_or_restart(cls, koji_instance: str, task_id: int, log_file_name: str):
        """Create a new koji task analysis"""
        with transaction(commit=True) as session:
            # Check if the task analysis already exists
            koji_task_analysis = (
                session.query(cls)
                .filter_by(koji_instance=koji_instance, task_id=task_id)
                .first()
            )
            if koji_task_analysis:
                # If it does, update the request_received_at timestamp
                koji_task_analysis.request_received_at = datetime.now(timezone.utc)
                session.add(koji_task_analysis)
                session.flush()
                return

            # If it doesn't, create a new one
            koji_task_analysis = KojiTaskAnalysis()
            koji_task_analysis.koji_instance = koji_instance
            koji_task_analysis.task_id = task_id
            koji_task_analysis.log_file_name = log_file_name
            session.add(koji_task_analysis)
            session.flush()

    @classmethod
    @backoff.on_exception(backoff.expo, OperationalError, max_tries=DB_MAX_RETRIES)
    def add_response(cls, task_id: int, metric_id: int):
        """Add a response to a koji task analysis"""
        with transaction(commit=True) as session:
            koji_task_analysis = session.query(cls).filter_by(task_id=task_id).first()
            # Ensure that the task analysis doesn't already have a response
            if koji_task_analysis.response:
                # This is probably due to an analysis that took so long that
                # a follow-up analysis was started before this one completed.
                # We want to maintain consistency between the response we
                # returned to the consumer, so we'll just drop this extra one
                # on the floor and keep the one saved in the database.
                return

            metric = (
                session.query(AnalyzeRequestMetrics).filter_by(id=metric_id).first()
            )
            koji_task_analysis.response = metric
            session.add(koji_task_analysis)
            session.flush()

    @classmethod
    @backoff.on_exception(backoff.expo, OperationalError, max_tries=DB_MAX_RETRIES)
    def get_response_by_task_id(cls, task_id: int) -> KojiStagedResponse:
        """Get a koji task analysis by task id"""
        with transaction(commit=False) as session:
            koji_task_analysis = session.query(cls).filter_by(task_id=task_id).first()
            if not koji_task_analysis:
                raise KojiTaskNotFoundError(f"Task {task_id} not yet analyzed")

            if not koji_task_analysis.response:
                # Check if the task analysis has timed out
                if koji_task_analysis.request_received_at.replace(
                    tzinfo=timezone.utc
                ) + timedelta(
                    minutes=SERVER_CONFIG.koji.analysis_timeout
                ) < datetime.now(timezone.utc):
                    raise KojiTaskAnalysisTimeoutError(
                        f"Task {task_id} analysis has timed out"
                    )

                # Task analysis is still in progress, so we need to let the
                # consumer know
                raise KojiTaskNotAnalyzedError(
                    f"Task {task_id} analysis is still in progress"
                )

            # We need to decompress the response message and return it
            response = LLMResponseCompressor.unzip(
                koji_task_analysis.response.compressed_response
            )
            return KojiStagedResponse(
                task_id=task_id,
                log_file_name=koji_task_analysis.log_file_name,
                response=response,
            )
